/* xfile.c
**
** Handle folder files with index files
**
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <fcntl.h>
#include <stdarg.h>
#include <errno.h>
#include <ctype.h>

#include "xlock.h"
#include "xfile.h"
#include "tools.h"
#include "mrmacs.h"
#include "bcv.h"
#include "vxerror.h"

#ifndef __USE_XOPEN_EXTENDED
#error __USE_XOPEN_EXTENDED is undefined
#endif

#define IXSFX ".idx"

#ifndef __GNUC__
#define __inline__
#endif

#define OFFSIZE (sizeof (off_t))

/* Was ist zu tun?
** Ich bentige
** a) Schreiboperationen, die ausschlielich an eine vorhandene
**    Folderdatei anhngen oder eine neue (cum Indexdatei) erzeugen.
** b) Leseoperationen, die ber einen Index die Position bestimmen, an der
**    sie in der Folderdatei lesen und am Ende des Eintrages ein EOF liefern.
*/
struct __xFile {
    int dfd, xfd;
    int mset;
    char mode[2];
    int msgno, new;
    off_t pos, xpos;
    /* `len' enthlt die Lnge der durch `xSeek' adressierten E-Mail */
    size_t len;
    /* `buf' zeigt IMMER auf den Anfang des Puffers; `bfp' zeigt auf das
     * nchste gltige Zeichen im Puffer; `bfe' zeigt (ausschlielich bei
     * Leseoperationen) auf das Ende der tatschlich im Puffer enthaltenen
     * Zeichen.
     */
    off_t ixlen, ixstart;
    char *buf, *bfp, *bfe;
    /* `bfs' enthlt die Lnge des Puffers, nicht die Anzahl der noch
     * enthaltenen Zeichen
     */
    char *extra, *prefix;
    int pfx_saved;
    size_t bfs;
};

static size_t BUFSIZE = 4096;

/* Die folgenden zwei Funktionen bentige ich an vielen Stellen innerhalb von
** `xfile.c'. Sie dienen dazu, einen Wert vom Typ `off_t' in
** `big endian byte order' aus einem Speicherbereich zu lesen bzw. in diesen
** zu schreiben; sie folgen keinerlei `alignment'-Regeln, d.h., die zu lesen-
** den (zu schreibenden) Werte knnen auch an ungeraden Adressen stehen.
*/
DEFCVF(b_to_off, off_to_b, off_t)

/* xoffset(xfd, n, indexlen, prefixlen, xposptr)
**     liefert die Dateiposition der n. Nachricht.
**     Fr n = 0 mu dies immer 0 sein.
**     Falls xposptr != NULL ist wird dort die Position des Indexeintrages
**     in der Indexdatei zurckgegeben.
*/
static off_t xoffset (int xfd, int pos, off_t ixlen, off_t ixstart, off_t *xpp)
{
    char pbf[OFFSIZE];
    off_t res = 0, fptr, flen = lseek (xfd, 0, SEEK_END);
    memset (pbf, '\0', OFFSIZE);
    if (pos < 0) return (off_t) -1;
    if (pos == 0) {
	if (xpp != (off_t *) NULL) { *xpp = ixstart; }
	return (off_t) 0;
    }
    fptr = (off_t) (pos - 1) * ixlen + ixstart;
    if (fptr >= flen) { return (off_t) -1; }
    if (xpp != (off_t *) NULL) { *xpp = fptr + ixlen; }
    if (lseek (xfd, fptr, SEEK_SET) < 0) { return (off_t) -1; }
    if (read (xfd, pbf, OFFSIZE) < 0) { return (off_t) -1; }
    res = b_to_off (pbf);
    return res;
}

/* xlength(fd, n)
 *     liefert die Lnge der n. Nachricht
 */
static ssize_t xlength (int xfd, int pos, off_t ixlen, off_t ixstart)
{
    off_t ep, sp;
    sp = xoffset (xfd, pos, ixlen, ixstart, (off_t *) NULL);
    ep = xoffset (xfd, pos + 1, ixlen, ixstart, (off_t *) NULL);
    if (sp < 0 || ep < 0) {
	return (ssize_t) -1;
    }
    return (ssize_t) (ep - sp);
}

/* Set the buffer size for the read/write buffer */
void xSetBufSize (size_t siz)
{
    BUFSIZE = siz;
}

static __inline__ int read_lenstart (int xfd, off_t *xlp, off_t *xsp)
{
    char buf[2 * OFFSIZE];
    off_t ixlen = (off_t) sizeof(off_t), ixstart = (off_t) (2*sizeof(off_t));
    lseek (xfd, 0, SEEK_SET);
    memset (buf, '\0', 2 * OFFSIZE);
    if (read (xfd, buf, 2 * OFFSIZE) != 2 * OFFSIZE) {
	*xlp = ixlen; *xsp = ixstart;
	return -1;
    } else {
	ixlen = b_to_off (buf);
	ixstart = b_to_off (&buf[OFFSIZE]);
	*xlp = ixlen; *xsp = ixstart;
	return 0;
    }
}

static __inline__ void set_pos (void *buf, off_t pos)
{
    off_to_b (pos, buf);
}

/* xOpen
** ffne die Folderdatei(en) zusammen mit der zugehrigen Indexdatei
** Es gibt folgende Arten diese Datei zu ffnen:
** a) im Schreibmodus (*amode == 'w'), in dem lediglich Daten an die vorhandene
**    Datei angehngt werden knnen (die Datei kann dabei auch neu erzeugt
**    worden sein),
** b) im Lesemodus (*amode == 'r'), in dem wahlfreier Zugriff, jedoch nur lesend
**    mglich ist.
** c) im Update-Modus (*amode == 'u'), in dem die Folderdatei(en) neu erzeugt
**    bzw. entleert werden; danach funktioniert dieser Modus genau wie der
**    Schreibmodus.
** d) im Index-Modus (*amode == 'i'), in dem zu einer vorhandenen Folderdatei
**    eine Indexdatei angelegt wird.
** Das zweite Zeichen von amode (amode[1]) gibt an, ob und wie eine Dateisperre
** fr die Datei angelegt werden soll:
** a) amode[1] == 's': es wird ein `shared lock' auf die gesamte Folderdatei
**    (und die zugehrige Indexdatei) gelegt; dies bedeutet, da mehrere
**    unterschiedliche Prozesse gleichzeitig auf diese Datei(en) zugreifen
**    knnen - jedoch keiner exklusiv.
**    Ist der Folder bereits mit einem `exclusive lock' belegt, dann wird der
**    Aufrufer schlafen gelegt, bis der `exclusive lock' aufgehoben wurde...
** b) amode[1] == 'x': es wird ein `exclusive lock' auf die gesamte Folderdatei
**    (und die zugehrige Indexdatei) gelegt; in diesem Modus kann genau ein
**    Proze auf diese Datei(en) zugreifen kann.
**    Ist der Folder bereits mit einem `shared lock' oder einem `exclusive lock'
**    belegt, dann wird der Aufrufer schlafen gelegt, bis alle entsprechenden
**    Locks wieder aufgehoben wurden...
** c) amode[1] == 'S': die Folderdatei wird - wie unter a) mit einem
**    `shared lock' belegt; allerdings kehrt `xOpen' mit einer Fehlermeldung
**    zurck, wenn der Folder bereits mit einem `exclusive lock' belegt wurde.
** d) amode[1] == 'X': die Folderdatei wird - wie unter b) mit einem
**    `exclusive lock' belegt; allerdings kehrt `xOpen' mit einer Fehlermeldung
**    zurck, wenn der Folder bereits mit einem `shared lock' oder
**    `exclusive lock' belegt wurde.
*/
xFile xOpen (char *path, char *amode, ...)
{
    va_list ap;
    int dfd, xfd;
    int mode = 0, oFlags, iFlags, lock, pf, lknb, lf;
    char *ixpath;
    off_t ixstart = 2 * OFFSIZE, ixlen = OFFSIZE;
    xFile res = (xFile) 0;
    va_start (ap, amode);
    /* Um zustzliche Informationen zu speichern (beispielsweise die
    ** UIDL fr POP3/IMAP4) bentige ich zustzliche Argumente. Diese
    ** Argumente sollen aber nur dann angegeben werden, wenn dies ausdrcklich
    ** gefordert ist; dazu mu ein zustzliches Flag angegeben werden: `e'
    */
    if (*amode != 'r' && noNULL(strchr (amode, 'e'))) {
	ixlen += va_arg (ap, off_t);
	ixstart += va_arg (ap, off_t);
    }
    /* Ich erlaube zwar nur lesenden Zugriff, ffne die Datei aber trotzdem
    ** lesend und schreibend. Der Grund hierfr ist einfach:
    ** Es gibt einige (vor allem ltere) Unix-Systeme, die ein Sperren von
    ** Dateien nur auf (auch) zum Schreiben geffneten Dateien zulassen.
    ** Da meine Routinen sich sowieso nach dem Mode-Flag (`i'...`u') richten,
    ** bleibt die Integritt der Mail-Folder trotzdem erhalten.
    */
    if (*amode == 'r') {
	/* Warum ich hier `RDWR' statt `RDONLY' setze?
	** Ich habe eine neue Funktion `xReWrite' eingefhrt, die (quasi) eine
	** Neuerffnung des Folders im Schreibmodus (mit vorheriger entleerung)
	** ermglicht. Damit ich nicht wirklich neu ffnen mu, mu der Folder
	** sowohl les- als auch schreibbar geffnet sein. Auf der Zugriffs-
	** schicht (ber die hier definierten Funktionen) durfte davon jedoch
	** nichts zu sehen sein.
	*/
	iFlags = oFlags = O_RDWR;
    } else if (*amode == 'w' || *amode == 'u') {
	iFlags = oFlags = O_CREAT|O_RDWR;
	mode = va_arg (ap, int);
    } else if (*amode == 'i') {
	oFlags = O_RDWR;
	iFlags = O_CREAT|O_TRUNC|O_RDWR;
	mode = va_arg (ap, int);
    } else {
	va_end (ap);
	return (xFile) 0;
    }
    /* ein `e'-Flag kann sowohl vor als auch hinter dem `s'- (bzw. `x'-)Flag
    ** stehen; es kann auch sein, da gar kein `s'- (`x'-)Flag angegeben wurde.
    ** Darum kann ich im folgenden nicht mehr davon ausgehen, da ein `s'-
    ** (`x'-)Flag an fester Position steht.
    */
    pf = (amode[1] == 'e' ? 2 : 1);
    va_end (ap);
    /* Ich ziehe das File locking aus dem Zugriffsmodus heraus und lege es
     * mittels eines weiteren Zeichens in der Zeichenkette `amode' fest;
     * dies hat den Vorteil, da dieses Modul universeller einsetzbar ist.
     * Natrlich mssen `vxdeliver', ... entsprechend angepat werden.
     */
    lf = tolower (amode[pf]);
    lock = (lf == 's'? XL_LKSH : (lf == 'x'? XL_LKEX : -1));
    lknb = isupper (amode[pf]) ? XL_LKNB : 0;
    if (!(ixpath = (char *) malloc (strlen (path) + strlen (IXSFX) + 1))) {
	return (xFile) 0;
    }
    strcpy (ixpath, path);
    strcat (ixpath, IXSFX);
    ifnull (res = (xFile) malloc (sizeof (struct __xFile) +
				  (size_t) (ixlen+ixstart))) {
	free (ixpath);
	return res;
    }
    res->ixlen = ixlen; res->ixstart = ixstart;
    res->extra = (char *)res + sizeof (struct __xFile);
    res->prefix = res->extra + ixlen;
    if ((dfd = open (path, oFlags, mode)) < 0) {
	free (res); free (ixpath);
	return (xFile) 0;
    }
    /* Ich habe endlich meine eigene Abstraktion (von diesen dmlichem system-
    ** abhngigen LOCK-Funktionen) geschrieben. (Diese ist zwar auch system-
    ** abhngig, hat jedoch - im Gegensatz zu allen anderen Funktionen dieser
    ** Art - eine einheitliche Schnittstelle. Auerdem ist sie, wenn sie
    ** ber `fcntl' definiert ist, auch flexibelsten von allen Varianten ein-
    ** setzbar.
    */
    if (lock >= 0) {
	if (xLock (dfd, lock|lknb, 0))
	{
	    close (dfd);
	    free (res); free (ixpath);
	    errno = EBUSY;
	    return (xFile) 0;
	}
    }
    if ((xfd = open (ixpath, iFlags, mode)) < 0) {
	close (dfd);
	free (res); free (ixpath);
	return (xFile) 0;
    }
    if (lock >= 0) {
	if (xLock (xfd, lock|lknb, 0) < 0)
	{
	    close (xfd); close (dfd);
	    free (res); free (ixpath);
	    errno = EBUSY;
	    return (xFile) 0;
	}
    }
    if (*amode == 'u') {
	/* I didn't open the file(s) above with O_TRUNC, because O_TRUNC
	** even works with an active lock on a file; i instead use `ftruncate'
	** for truncating the folder/index file to a zero length.
	*/
	ftruncate (dfd, (off_t) 0);
	ftruncate (xfd, (off_t) 0);
    }
    /* Now, even files opened with `u' are not simply overwritten if they
    ** are locked by another process
    */
    res->mode[0] = *amode;
    res->mode[1] = lf;
    res->dfd = dfd; res->xfd = xfd;
    res->new = 0;
    if (*amode == 'r') {
	/* Bei einem Lesezugriff mssen die Lngen- bzw. Startinformation der
	** Indexeintrge aus der (vorhandenen!) Indexdatei gelesen werden.
	*/
	xFile res1;
	read_lenstart (xfd, &ixlen, &ixstart);
	/* Anschlieend mu die xFile-structure in ihrer Lnge entsprechend
	** angepat werden.
	*/
	ifnull (res1 = (xFile) realloc (res, sizeof (struct __xFile) +
					     (size_t) (ixlen + ixstart))) {
	    close (dfd); close (xfd); return (xFile) 0;
	}
	res = res1;
	/* Die Anpassung der Lngen- uns Startinformation sowie der Zeiger
	** auf die zustzlichen Speicherbereiche mu ebenfalls erfolgen.
	*/
	res->ixlen = ixlen; res->ixstart = ixstart;
	res->extra = (char *)res + sizeof (struct __xFile);
	res->prefix = res->extra + ixlen;
	/* Bei einem Lesezugriff mu nach `open' die Ausgangsposition
	** zunchst die Position der ersten Nachricht sein. Auerdem
	** mssen die Zusatzinformationen (Prefix + Extra) gelesen werden.
	*/
	lseek (xfd, 0, SEEK_SET);
	read (xfd, res->prefix, res->ixstart);
	lseek (xfd, res->ixstart, SEEK_SET);
	read (xfd, res->extra, res->ixlen);
	res->pos = xoffset (xfd, 0, res->ixlen, res->ixstart, &res->xpos);
	res->len = xlength (xfd, 0, res->ixlen, res->ixstart);
    } else if (*amode == 'i') {
	/* Wenn der Folder im Index-Modus geffnet wurde, dann mu das Flag,
	** welches angibt, da eine Nachricht zu ende ist, stets ungesetzt
	** sein (und bleiben).
	*/
	res->mset = 0;
	res->pos = (off_t) 0;
	res->len = (size_t) lseek (dfd, 0, SEEK_END);
	lseek (dfd, 0, SEEK_SET);
	memset (res->extra, '\0', ixlen);
	memset (res->prefix, '\0', ixstart);
	off_to_b (ixlen, res->prefix);
	off_to_b (ixstart, res->prefix + OFFSIZE);
	write (xfd, res->prefix, ixstart);
	res->xpos = lseek (xfd, 0, SEEK_CUR);
	res->new = 1;
    } else {
	/* Bei einem Schreib- oder Updatezugriff mu das Flag, welches angibt,
	** da eine Nachricht zu ende ist, zunchst gesetzt sein; wird nmlich
	** nichts geschrieben, dann mu die Datei unverndert bleiben; wird
	** hingegen etwas geschrieben, dann mu zum Schlu ein neuer Marker
	** fr die Dateiposition am Ende der Nachricht (`xSet' oder `xClose')
	** eingetragen werden. Dies darf aber nur dann geschehen, wenn
	** tatschlich etwas geschrieben wurde.
	*/
	off_t xfend = lseek (xfd, 0, SEEK_END);
	if (xfend > (off_t) 0) {
	    /* Wenn bei einem Schreibzugriff die Lnge der Indexdatei > 0 ist,
	    ** d.h., wenn die Indexdatei schon vorher existiert hat, dann
	    ** mssen (wie bei einem Lesezugriff!) die Lngen- bzw. Start-
	    ** information der Indexeintrge aus der Indexdatei gelesen werden.
	    */
	    xFile res1;
	    read_lenstart (xfd, &ixlen, &ixstart);
	    /* Anschlieend mu die xFile-structure in ihrer Lnge entsprechend
	    ** angepat werden.
	    */
	    res1 = (xFile) realloc (res, sizeof (struct __xFile) +
					 (size_t) (ixlen + ixstart));
	    if (res1 == NULL) {
		close (dfd); close (xfd); return (xFile) 0;
	    }
	    res = res1;
	    /* Die Anpassung der Lngen- uns Startinformation sowie der Zeiger
	    ** auf die zustzlichen Speicherbereiche mu ebenfalls erfolgen.
	    */
	    res->ixlen = ixlen; res->ixstart = ixstart;
	    res->extra = (char *)res + sizeof (struct __xFile);
	    res->prefix = res->extra + ixlen;
	    /* Anschlieend mu die Prefix-Information gelesen werden. */
	    lseek (xfd, 0, SEEK_SET);
	    read (xfd, res->prefix, res->ixstart);
	} else {
	    /* In diesem Fall liegt (eindeutig!) eine neue oder eine geleerte
	    ** Indexdatei vor.
	    */
	    memset (res->extra, '\0', ixlen);
	    memset (res->prefix, '\0', ixstart);
	    off_to_b (ixlen, res->prefix);
	    off_to_b (ixstart, res->prefix + OFFSIZE);
	    write (xfd, res->prefix, ixstart);
	    res->new = 1;
	}
	res->mset = 1;
	/* Ebenso mu (bei einem Schreibzugriff) der Dateizeiger ans Ende der
	** Datei Positioniert werden. Die Zusatzinformationen sind schon
	** geschrieben (bei einer neuen (oder entleerten) Indexdatei), so da
	** auch in diesem Fall der Dateizeiger auf der richtigen Position
	** landet.
	*/
	res->pos = lseek (dfd, 0, SEEK_END);
	res->xpos = lseek (xfd, 0, SEEK_END);
    }
    res->msgno = (int) ((res->xpos - res->ixstart) / res->ixlen);
    res->buf = (char *) 0;
    res->bfp = (char *) 0;
    res->bfe = (char *) 0;
    res->pfx_saved = 1;
    free (ixpath);
    return res;
}


/* Setze den internen Dateizeiger auf den Anfang der Nachricht `index' (von
 * Null ab gezhlt!) und die interne Lnge auf die Lnge der Nachricht.
 * Diese Operation funktioniert nur bei lesendem Zugriff auf die Datei.
 * (Warum die letzte Einschrnkung?
 *  Nun, ich gehe davon aus, da beim Eintreffen einer neuen Mail, diese immer
 *  hinten an den Folder angehngt wird; Schreibzugriff bedeutet hier also
 *  einen rein sequentiellen Zugriff, whrend man bei einem Lesezugriff einen
 *  Indexsequentiellen Zugriff mittels der Nummer der Nachricht als Index hat.
 *  Das sind dann genau die Operationen, die fr einen normalen Zugriff ben-
 *  tigt werden (SMTP - Schreibzugriff, POP3 - Lesezugriff).
 *  Der fr das Entfernen einzelner Messages notwendigen Zugriff mu sowieso
 *  auf eine ganz andere Art geschehen, so da ich hierbei verbleiben kann.)
 */
int xSeek (xFile fx, off_t index)
{
    off_t offset, xpos = (off_t) 0;
    ssize_t size;
    /* Vergewissere dich, da `fx' mit Lesezugriff geffnet wurde. */
    if (fx->mode[0] != 'r') { errno = EACCES; return -1; }
    /* Wenn der angegebene Index auerhalb des zulssigen Bereiches liegt,
     * dann liefert `xoffset' einen negativen Wert; kehre dann ebenfalls mit
     * einem negative Wert zurck.
     */
    if ((offset = xoffset (fx->xfd, index, fx->ixlen, fx->ixstart, &xpos)) < 0)
    {
	return -2;
    }
    fx->xpos = xpos;
    /* Berechne die Lnge der Nachricht aus zwei aufeinanderfolgenden Indices
     * Wenn diese < 0 ist, dann */
    if ((size = xlength (fx->xfd, index, fx->ixlen, fx->ixstart)) < 0) {
	return -3;
    }
    /* Trage die Werte fr die Dateiposition und die Message-Lnge in die
     * internen Variablen der Struktur ein.
     */
    fx->pos = offset;
    fx->len = size;
    fx->msgno = (int) index;
    /* Setze den Dateizeiger der Datendatei auf den Anfang der zugehrigen
     * E-Mail innerhalb der Datendatei.
     */
    offset = lseek (fx->dfd, fx->pos, SEEK_SET);
    if (fx->buf) {
	fx->bfp = fx->bfe = fx->buf;
    }
    lseek (fx->xfd, xpos, SEEK_SET);
    read (fx->xfd, fx->extra, fx->ixlen);
    return index;
}

static __inline__ int bfalloc (xFile fx)
{
    ifnull (fx->buf) {
	ifnull (fx->buf = (char *) malloc (BUFSIZE)) {
	    return -1;
	}
	fx->bfs = BUFSIZE;
	/* Aktuelle Schreibposition */
	fx->bfp = fx->buf;
	/* Ende der eingelesenen Daten */
	fx->bfe = fx->buf;
    }
    return 0;
}

ssize_t xFlush (xFile fx)
{
    ssize_t res;
    if (fx->bfp == fx->buf) {
	return 0;
    }
    res = write (fx->dfd, fx->buf, (size_t) (fx->bfp - fx->buf));
    fx->bfp = fx->buf;
    return res;
}

ssize_t xWrite (xFile fx, void *buf, size_t bfs)
{
    size_t rsz;
    ssize_t res = 0, r1;
    /* Wenn dei Datei weder fr Schreib- noch fr Update-Modus geffnet wurde,
    ** dann darf keine Schreiboperation ausgefhrt werden.
    */
    unlessnull (strchr ("ri", fx->mode[0])) { errno = EACCES; return -1; }
    if (bfalloc (fx) < 0) { return -1; }
    /* Wenn der Puffer angelegt wurde, dann mu(!) das Marker-Flag
     * zurckgesetzt werden, da dies schon als abgeschlossene Schreiboperation
     * gilt (bzw. betrachtet werden mu); es wird ja nicht sofort geschrieben,
     * sondern zeitversetzt - nmlich dann, wenn der Puffer gefllt ist.
     */
    fx->mset = 0;
    /* Restlnge des Puffers berechnen */
    rsz = fx->bfs - (fx->bfp - fx->buf);
    /* Solange noch Daten vorhanden sind */
    while (bfs > rsz) {
	/* bertrage Daten in den Schreibpuffer bis dieser gefllt ist. */
	memcpy (fx->bfp, buf, rsz);
	fx->bfp += rsz;
	/* Wenn der Puffer gefllt ist, dann bertrage seinen Inhalt in
	 * die Datendatei
	 */
	if ((r1 = xFlush (fx)) < 0) { return res; }
	/* Die Anzahl der bisher bertragenen Bytes geht nach `res'. */
	res += r1;
	/* Setze den Eingabedatenzeiger auf die nchsten Eingabedaten */
	buf =  (char *) buf + rsz;
	/* Verringere die Eingabedatenlnge um die Anzahl der bereits in
	 * die Datendatei bertragenen Bytes
	 */
	bfs -= rsz;
	/* Setze die Restlnge des Puffers neu */
	rsz = fx->bfs;
    }
    /* Wenn noch Daten brig sind, dann bertragge sie in den Puffer */
    if (bfs > 0) {
	memcpy (fx->bfp, buf, bfs);
	fx->bfp += bfs;
	res += bfs;
    }
    return res;
}

/* Flle den Puffer neu; diese Operation wird NUR dann ausgefhrt, wenn
 * der Puffer vollstndig leer ist.
 */
static ssize_t xfill (xFile fx)
{
    ssize_t rlen;
    ssize_t rest;
    rest = fx->len - (size_t) (lseek (fx->dfd, 0, SEEK_CUR) - fx->pos);
    /* Wenn keine Bytes in der Nachricht mehr brig sind, dann gib 0 zurck */
    if (rest <= 0) return 0;
    /* Wenn noch mehr Daten in der Nachricht sind als in den Puffer passen,
     * dann lies nur soviele von diesen Daten, wie in den Puffer passen.
     */
    if (rest > fx->bfs) rest = (ssize_t) fx->bfs;
    /* Wenn die Leseoperation fehlschlgt, dann gib -1 zurck. */
    if ((rlen = read (fx->dfd, fx->buf, (size_t) rest)) < 0) { return -1; }
    /* Wenn weniger Bytes gelesen wurden als erwartet, dann setze die
     * Restlnge (das Funktionsergebnis) so, da sie der Anzahl der
     * tatschlich gelesenen Bytes entspricht.
     */
    if (rlen < rest) rest = rlen;
    /* Setze den aktuellen Lesezeiger und das aktuelle Pufferende */
    fx->bfp = fx->buf;
    fx->bfe = fx->buf + rest;
    /* Kehre mit der Anzahl der tatschlich gelesenen Bytes zurck */
    return rest;
}

/* Fr die Leseoperation bentige ich einen weiteren Zeiger;
 * auerdem sollen nur noch soviele Daten bertragen werden, wie noch in der
 * Nachricht enthalten sind
 */
ssize_t xRead (xFile fx, void *buf, size_t bfs)
{
    size_t rsz;
    ssize_t res = 0;
    ssize_t rblen;
    /* berprfe, ob lesen mglich ist. */
    unlessnull (strchr ("wu", fx->mode[0])) { errno = EACCES; return -1; }
    /* Wenn kein Speicherplatz mehr brig fr den Puffer, dann kann auch
     * nichts mehr gelesen werden.
     */
    if (bfalloc (fx) < 0) { return -1; }
    /* Wenn noch Daten im Puffer sind, dann bertrage zunchst diese */
    rsz = fx->bfe - fx->bfp;
    if (rsz > 0) {
	if (rsz > bfs) { rsz = bfs; }
	memcpy (buf, fx->bfp, rsz);
	fx->bfp += rsz;
	buf = (char *) buf + rsz;
	bfs -= rsz;
	res += (ssize_t) rsz;
    }
    /* Solange noch Daten gelesen werden mssen */
    while (bfs > 0) {
	/* Wenn keine Daten mehr brig sind oder ein Fehler aufgetreten ist,
	 * dann beende und kehre mit der Anzahl der bisher gelesenen Daten
	 * zurck.
	 */
	if ((rblen = xfill (fx)) <= 0) { return res; }
	/* Berechne die Menge der Daten im Puffer */
	rsz = fx->bfe - fx->bfp;
	/* Wenn mehr Daten im Puffer sind, als bentigt werden, dann
	 * bertrage nur soviele Daten, wie bentigt werden.
	 */
	if (rsz > bfs) { rsz = bfs; }
	memcpy (buf, fx->bfp, (size_t) rsz);
	/* Setze den Pufferzeiger entsprechend weiter */
	fx->bfp += rsz;
	/* Setze den Ergebniszeiger entsprechend weiter */
	buf = (char *) buf + rsz;
	/* Berechne die Menge der noch zu lesenden Daten */
	bfs -= rsz;
	res += rsz;
    }
    return res;
}

ssize_t xMsgLen (xFile fx, int pos)
{
    if (fx->mode[0] != 'r') { errno = EACCES; return -1; }
    if (pos < 0) {
	return fx->len;
    }
    return xlength (fx->xfd, pos, fx->ixlen, fx->ixstart);
}

/* Trage eine Zeiger auf die aktuelle Dateiposition in der Datendatei in die
** Indexdatei ein.
** Wenn der Folder im Index-Modus (s.o.) geffnet wurde, dann mu stattdessen
** die gewnschte Dateiposition als zweiter Parameter angegeben werden - da
** dann von einer schon existierenden Folderdatei ohne Index ausgegangen
** werden mu.
*/
int xSet (xFile fx, ...)
{
    off_t fofs;
    va_list av;
    /* Diese Operation darf in allen Modi auer dem Lese-Modus erfolgen. */
    if (fx->mode[0] == 'r') { errno = EACCES; return -1; }
    /* Setze nur dann einen Marker, wenn noch keiner vorhanden ist! */
    if (fx->mset) { return 0; }
    /* Wenn die Datei im Index-Modus geffnet wurde, dann mu die
    ** einzustellende Indexposition als zweites Argument bergeben werden;
    ** wurde sie hingegen im Schreib- oder Update-Modus geffnet, dann ist
    ** die Indexposition der aktuelle Stand des Dateizeigers nach einer
    ** erfolgten `flush'-Operation.
    */
    if (fx->mode[0] == 'i') {
	va_start (av, fx);
	fofs = va_arg (av, off_t);
	va_end (ap);
    } else {
	xFlush (fx);
	fofs = lseek (fx->dfd, 0, SEEK_CUR);
    }
    /* Die Dateipositionen speichere ich grundstzlich in `big endian order'
    ** ab. So gibt es keinerlei Probleme, das ganze auf Architekturen mit
    ** unterschiedlicher Byteorder zu verwenden.
    */
    /* Wandle die aktuelle Dateiposition in eine MSB-geordnete Bytefolge um */
    off_to_b (fofs, fx->extra);
    /* Schreibe die eben erzeugte Bytefolge (zusammen mit den brigen
    ** Zusatzinformationen) in die Indexdatei
    */
    if (write (fx->xfd, fx->extra, fx->ixlen) != fx->ixlen) {
	return (off_t) -1;
    }
    /* Schreibe - falls dies notwendig ist - die Prefix-Daten ebenfalls
    ** zurck.
    */
    if (!fx->pfx_saved) {
	fofs = lseek (fx->xfd, 0, SEEK_CUR);
	lseek (fx->xfd, 0, SEEK_SET);
	off_to_b (fx->ixlen, fx->prefix);
	off_to_b (fx->ixstart, fx->prefix + OFFSIZE);
	write (fx->xfd, fx->prefix, fx->ixstart);
	lseek (fx->xfd, fofs, SEEK_SET);
    }
    /* Im Index-Modus darf (selbstverstndlich das Setzen einer Indexposition
    ** niemals unterbunden werden; dieser Modus dient ja gerade zum Anlegen
    ** der Indexdatei. In den beiden anderen Fllen (Write- und Update-Modus)
    ** hingegen wird durch dieses Flag das Setzen einer ungltigen Indexposi-
    ** tion (keine weiteren Daten geschrieben) unterbunden. Das Setzen eines
    ** Index wird durch die erste `write'-Operation wieder ermglicht.
    */
    if (fx->mode[0] != 'i') fx->mset = 1;
    fx->new = 0;
    return 0;
}

/* Setze den Lesezeiger auf den Anfang der aktuellen Meldung zurck oder
** beende den Indexierungsmodus (`i').
** Wie gut, da ich diese Routine schon eingeplant hatte; ich bentige sie
** fr das Rcksetzen nach einer automatischen Erkennung einer komprimierten
** E-Mail, sowie fr die Umschaltung vom Schreib-, Update- und Indexierungs-
** in den Lesemodus.
*/
int xReSet (xFile fx)
{
    /* Also:
    **   Wenn der Folder im Schreib-, Update- oder Indexmodus geffnet wurde,
    **   dann soll er nun im Lesemodus geffnet sein, wobei dann die internen
    **   Zeiger auf die erste im Folder enthaltene E-Mail zurckgesetzt werden.
    */
    unlessnull (strchr ("iuw", fx->mode[0])) {
	if (fx->mode[0] != 'i') { xSet (fx); }
	fx->mode[0] = 'r';
	if (xSeek (fx, 0) < 0) { return -1; }
	return 0;
    } else {
	/* Ist der Folder hingegen im Lesemodus geffnet, dann werden die
	** internen Zeiger auf den Anfang der aktuellen E-Mail zurckgestellt.
	** und die Zustzlichen Daten neu eingelesen.
	*/
	if (lseek (fx->dfd, fx->pos, SEEK_SET) < 0) { return -1; }
	if (lseek (fx->xfd, fx->xpos, SEEK_SET) < 0) { return -1; }
	if (read (fx->xfd, fx->extra, fx->ixlen) != fx->ixlen) { return -1; }
    }
    /* Diese Stelle hatte ich vergessen; natrlich mu (!!!) der Eingabepuffer
    ** nach einer Re-Positionierung auf den Anfang der E-Mail geleert werden,
    ** da sonst Zeichen beim nchsten Einlesevorgang doppelt gelesen werden
    ** knnten!
    */
    fx->bfp = fx->bfe = fx->buf;
    return 0;
}

int xSetWrite (xFile fx)
{
    /* Also:
    **   Wenn der Folder bereits im Schreib, Update- oder Indexmodus geffnet
    **   wurde, dann soll gar nichts geschehen; ist er hingegen im Lesemodus
    **   geffnet, dann soll er nun in den Schreibmodus gebracht werden - wobei
    **   dann auch der Dateizeiger ans Ende der Datei gesetzt werden mu.
    */
    unlessnull (strchr ("iuw", fx->mode[0])) { return -1; }
    if (lseek (fx->dfd, (off_t) 0, SEEK_END) < 0) { return -1; }
    if (lseek (fx->xfd, (off_t) 0, SEEK_END) < 0) { return -1; }
    if (read (fx->xfd, fx->extra, fx->ixlen) != fx->ixlen) { return -1; }
    fx->bfp = fx->bfe = fx->buf;
    fx->mode[0] = 'w';
    return 0;
}

/* Schneide die `xFile'-Dateien ab den dem angegebenen Index entsprechenden Positionen
** ab. Diese Positionen werden durch einen Aufruf von `xoffset' bestimmt.
** Auerdem werden die `xFile'-internen Dateizeiger auf die dem Index entsprechenden
** Positionen zurckgesetzt (mittels `xSeek').
*/
int xTruncate (xFile fx, int index)
{
    char mode = fx->mode[0];
    off_t xoff, doff;
    /* Abschneiden ist nur! in den beiden Schreib-Modi mglich... */
    if (mode != 'w' && mode != 'u') { return -1; }
    /* Im ersten Schritt mssen die noch im Puffer stehenden Daten in die Datendatei
    ** bertragen werden, damit in jedem Fall hinterher ein konsistenter Zustand in
    ** der `xFile'-Struktur steht.
    */
    xFlush (fx);
    if (index < 0) { index = fx->msgno; }
    xoff = fx->ixstart + index * fx->ixlen;
    doff = xoffset (fx->xfd, index, fx->ixlen, fx->ixstart, &xoff);
    if (doff < (off_t) 0) { return -1; }
    /* An dieser Stelle kann ich sagen:
    ** a) beide Dateien - Index und Datendatei - sind schreibbar geffnet,
    ** b) beides sind gltige regulre Dateien,
    ** c) beide Dateideskriptoren sind gltig.
    ** Wren diese Angaben nicht erfllt, dann wre das ganze sowieso nicht mehr
    ** konsistent, so da an dieser Stelle auch durch das unbedingte Ausfhren der
    ** beiden `ftruncate'-Aufrufe ohne Fehlerrckgabe nicht noch mehr Schaden entstehen
    ** kann, als sowieso schon entstanden ist.
    ** Sind andererseits die oben genannten Annahmen erfllt, dann werden beide Aufrufe
    ** von `ftruncate' mit 0 (als Funktionsergebnis) zurckkehren, so da ich dieses
    ** Ergebnis dann sowieso nicht mehr bentige.
    */
    (void) ftruncate (fx->dfd, doff);
    (void) ftruncate (fx->xfd, xoff);
    /* Nachbearbeitung: Die entsprechenden Daten in der `xFile'-Datenstruktur mssen
    ** korrigiert werden.
    */
    (void) xSeek (fx, index);
    return 0;
}

int xReWrite (xFile fx)
{
    /* Wenn der Folder im Indexmodus geffnet wurde, dann darf kein
    ** `rewrite' auf ihn erfolgen.
    */
    if (fx->mode[0] == 'i') { errno = EACCES; return -1; }

    /* Andernfalls werden Folder- und Indexdatei
    ** (mittels `ftruncate') entleert und Lese- bzw. Schreibpuffer frei-
    ** gegeben.
    */
    ftruncate (fx->dfd, (off_t) 0); lseek (fx->dfd, 0, SEEK_SET);
    ftruncate (fx->xfd, (off_t) 0); lseek (fx->xfd, 0, SEEK_SET);
    free (fx->buf); fx->buf = fx->bfp = fx->bfe = NULL;
    /* Nun werden noch die Positions bzw. Lngenangabe zurckgesetzt sowie
    ** der Modus auf `u' (update) gesetzt - schlielich haben wir die Daten
    ** ja gerade eben gelscht ;-).
    */
    fx->pos = (off_t) 0; fx->len = (size_t) 0;
    fx->mode[0] = 'u';
    /* Trage zuletzt noch den Prefix in die Indexdatei ein. */
    memset (fx->extra, '\0', fx->ixlen);
    memset (fx->prefix, '\0', fx->ixstart);
    off_to_b (fx->ixlen, fx->prefix);
    off_to_b (fx->ixstart, fx->prefix + OFFSIZE);
    write (fx->xfd, fx->prefix, fx->ixstart);
    fx->xpos = lseek (fx->xfd, 0, SEEK_CUR);
    fx->mset = 1;
    return 0;
}

/* xAppend soll den Folder - hnlich wie xReWrite - wieder zum Schreiben
** ffnen; der Unterschied liegt lediglich darin, da die ursprnglichen
** Daten erhalten bleiben sollen...
*/
int xAppend (xFile fx)
{
    vxerror (EX_SOFTWARE, 0, "xAppend isn't implemented (yet)!\n");
    return 0;
}

/* xDelete entfernt die angegebene Liste von E-Mails aus dem Folder. Dabei
** werden alle anderen E-Mails entsprechend nach vorne geschoben - und ihre
** Positionsangaben innerhalb der Indexdatei entsprechend korrigiert...
** Das letzte Element der als Vektor bergebenen Liste sollte einen Wert < 0
** (vorzugsweise -1) enthalten...
*/
int xDelete (xFile fx, int *msgs)
{
    vxerror (EX_SOFTWARE, 0, "xDelete isn't implemented (yet)!\n");
#if 0
    {
	int ix;
	off_t oldpos, newpos;
	size_t msgsz, gapsz;
        res->pos = xoffset (xfd, 0, res->ixlen, res->ixstart, &res->xpos);
        res->len = xlength (xfd, 0, res->ixlen, res->ixstart);                  
	for (ix = 0; msgs[ix] >= 0; ++ix) {
	    fx->pos = xoffset (fx->xfd, 0, fx->ixlen, fx->ixstart, &oldpos);
    }
#endif
    return 0;
}

/* Bei einer Schlieoperation mu folgendes getan werden:
** a) Wenn die Datei zum Schreiben geffnet war, dann mu der Puffer zurck-
**    geschrieben werden.
** b) Wenn die Datei zum Schreiben geffnet war, dann mu die
**    Dateiendeposition in der Indexdatei abgelegt werden.
** c) Wenn ein Schreib- oder Lesepuffer angelegt wurde, dann mu dieser
**    freigegeben werden.
** d) Die Daten- und die Indexdatei mssen geschlossen werden.
** e) Das xFile-Handle mu freigegeben werden.
*/
int xClose (xFile fx)
{
    int e0 = 0, e1, e2;
    /* Wenn der Folder im Schreib- oder Update-Modus geffnet wurde,
    ** dann mu vor dem Schlieen eine xSet-Operation erfolgen.
    ** Da ich im Index-Modus sowieso die einzelnen Positionen von Hand setzen
    ** mu, wird dies von `xClose' nicht mehr bernommen.
    */
    unlessnull (strchr ("wu", fx->mode[0])) {
	/* a) und b) werden durch den Aufruf von xSet automatisch erledigt. */
	e0 = xSet (fx);
    }
    /* Schritt c) */
    unlessnull (fx->buf) { free (fx->buf); }
    lseek (fx->xfd, (off_t) 0, SEEK_SET);
    write (fx->xfd, fx->prefix, fx->ixstart);
    /* Schritt d) */
    if (fx->mode[1] == 's' || fx->mode[1] == 'x') {
	/* explicitely un-locking of the two files; some systems seem to
	** require a lot of time for doing this automatically.
	*/
	xLock (fx->dfd, XL_UNLK, (off_t) 0);
	xLock (fx->xfd, XL_UNLK, (off_t) 0);
    }
    e1 = close (fx->dfd);
    e2 = close (fx->xfd);
    /* Schritt e) */
    free (fx);
    return (e0 ? e0 : (e1 ? e1 : e2));
}

/* Zustzliche Funktionen */

/* xPuts(fx, string)
**     Schreibe eine '\0'-terminierte Zeichenkette in die Datei `fx'
*/
ssize_t xPuts (char *buf, xFile fx)
{
    return xWrite (fx, buf, strlen (buf));
}

/* xGetc(fx)
**     Lies ein einzelnes Zeichen aus der Datei `fx'
** Analog zu `getc' liefert `xGetc' eine ganze Zahl. Das ermglicht es,
** im Fehlerfall (oder auch bei Dateiende) einen passenden Wert zurckzugeben;
** Im Fehlerfall wird `-2', bei Dateiende `-1' zurckgegeben.
*/
int xGetc (xFile fx)
{
    ssize_t rsz, rblen;
    /* Weder im Schreib- noch im Update-Modus knnen Daten gelesen werden. */
    unlessnull (strchr ("wu", fx->mode[0])) { errno = EACCES; return -2; }
    /* Wenn kein Speicherplatz mehr brig fr den Puffer, dann kann auch
    ** nichts mehr gelesen werden.
    */
    if (bfalloc (fx) < 0) { return -2; }
    /* Wenn noch Daten im Puffer sind, dann bertrage zunchst diese */
    rsz = fx->bfe - fx->bfp;
    if (rsz <= 0) {
	rblen = xfill (fx);
	if (rblen < 0) return -2;
	if (rblen == 0) return -1;
    }
    return (*fx->bfp++ & 0xff);
}

/* xGets (buf, siz, fx)
**     Lies eine Zeile - hchstens aber `siz' Zeichen aus der Datei`fx'
**     in den Puffer `buf'.
** Im Unterschied zu `fgets' wird hier nicht ein Zeiger auf eine Zeichenkette
** zurckgegeben, sondern die Anzahl der gelesenen Zeichen.
*/
ssize_t xGets (char *buf, size_t siz, xFile fx)
{
    ssize_t rsz, bsz, psz = 0;
    char *p = (char *) NULL;
    if (bfalloc (fx) < 0) { return -2; }
    /* Weder im Schreib- noch im Update-Modus knnen Daten gelesen werden. */
    unlessnull (strchr("wu", fx->mode[0])) { errno=EACCES; return (ssize_t)-1; }
    rsz = (ssize_t) siz - 1;
    while (isNULL(p) && rsz > 0) {
	/* Teste zuerst mal den Pufferinhalt */
	bsz = fx->bfe - fx->bfp;
	if (bsz <= 0 && xfill (fx) <= 0) { break; }
	/* Suche nach dem ersten Zeilenende im Puffer */
	p = memchr (fx->bfp, '\n', fx->bfe - fx->bfp);
	/* Wenn ein Zeilenende gefunden wurde, dann bertrage alles bis zu
	** diesem (einschlielich); ansonsten bertrage einfach den gesamten
	** Pufferinhalt.
	*/
	bsz = (noNULL(p) ? (p + 1) : fx->bfe) - fx->bfp;
	/* bertrage natrlich nur soviele Bytes, wie in den Zielbereich
	** hineinpassen.
	*/
	if (bsz > rsz) { bsz = rsz; }
	/* - und zwar von der aktuellen Position im Puffer an. */
	memcpy (buf, fx->bfp, bsz);
	/* Setze die Zeiger entsprechend weiter.*/
	fx->bfp += bsz; buf += bsz; rsz -= bsz;
	/* Die gesamte Anzahl der bisher gelesenen Bytes steht hiernach in
	** `psz'.
	*/
	psz += bsz;
    }
    /* Setze in jedem Fall noch ein NUL-Zeichen hinterher */
    *buf = '\0';
    return psz;
}

/* xPutc (cc, fx)
**     Schreibe ein einzelnes Zeichen in die Datei `fx'.
** Im Fehlerfall wird `-2' zurckgegeben, sonst `0'. Die Ausgabe erfolgt -
** genau wie bai allen anderen hier definierten Schreib- und Leseoperationen -
** gepuffert.
*/
int xPutc (int cc, xFile fx)
{
    ssize_t rsz;
    /* Weder im Lese- noch im Index-Modus drfen Daten geschrieben werden. */
    unlessnull (strchr ("ri", fx->mode[0])) { errno = EACCES; return -2; }
    /* Wenn kein Speicherplatz mehr fr den Puffer brig ist, dann kann auch
    ** nichts mehr gelesen werden.
    */
    if (bfalloc (fx) < 0) { return -2; }
    rsz = fx->bfs - (fx->bfp - fx->buf);
    if (rsz <= 0) { xFlush (fx); }
    *fx->bfp++ = (char) cc;
    return 0;
}

/* xUnlink (filename)
**     lscht den Folder `filename' (mit samt seinem Index).
** Wie die Indexdatei zu einem Folder heit soll ausschlielich durch dieses
** Modul bestimmt werden. Um das zu gewhrleisten, mute ich eine erweiterte
** Dateilschfunktion einfhren, die mit dieser Situation umgehen kann.
*/
int xUnlink (char *file)
{
    char *ixfile, e1, e2;
    ifnull (ixfile = (char *) malloc (strlen (file) + strlen (IXSFX) + 1)) {
	return -1;
    }
    strcpy (ixfile, file);
    strcat (ixfile, IXSFX);
    e1 = unlink (file);
    e2 = unlink (ixfile);
    free (ixfile);
    return (e1 ? e1 : e2);
}

/* xNoIndex (filename)
**     lscht die Indexdatei zu einem Folder.
** Diese Operation bentige ich fr die E-Mail-Folder eines Standardsystems
** (`smail' oder `sendmail').
** Damit mein POP-Server nmlich mit diesen umgehen kann, mu temporr eine
** Indexdatei erzeugt werden. Damit diese nach vollendeter POP3-Session aber
** wieder gelscht werden kann, bentige ich eine eigene Lschroutine.
*/
int xNoIndex (char *file)
{
    char *ixfile, e;
    ifnull (ixfile = (char *) malloc (strlen (file) + strlen (IXSFX) + 1)) {
	return -1;
    }
    strcpy (ixfile, file);
    strcat (ixfile, IXSFX);
    e = unlink (ixfile);
    free (ixfile);
    return e;
}

/* xEmpty (filehandle)
**     gibt 1 (true) zurck, falls der Folder, auf den gerade mittels
**     `filehandle' zugegriffen wird, leer ist und 0 (false), wenn nicht.
*/
int xEmpty (xFile fx)
{
    off_t oldoff = lseek (fx->dfd, 0, SEEK_CUR);
    off_t newoff = lseek (fx->dfd, 0, SEEK_END);
    if (oldoff < 0 || newoff < 0) { return -1; }
    lseek (fx->dfd, oldoff, SEEK_SET);
    if (newoff) { return 0; }
    return 1;
}

/* xIsNew (filehandle)
**     gibt 1 zurck, falls die Indexdatei vllig neu erzeugt wurde und
**     0 sonst. Sobald das erste mal ein xSet aufgerufen wurde, ist das
**     Ergebnis 0.
*/
int xIsNew (xFile fx)
{
    return fx->new;
}

/* xChown (filehandle, uid, gid)
**     setzt den Besitzer bzw. die Gruppe der Folder- und Indexdatei auf
**     die bergebenen Werte `uid' `gid'; damit das ganze funktioniert, mu
**     die `real uid' des Prozesses die des Superusers (uid = 0) sein.
*/
int xChown (xFile fx, uid_t uid, gid_t gid)
{
    int ec;
    uid_t euid = geteuid (), ruid = getuid();
    if (getuid () != (uid_t) 0) { errno = EPERM; return -1; }
    setreuid (ruid, ruid);
    ec = fchown (fx->dfd, uid, gid);
    if (ec == 0) { ec = fchown (fx->xfd, uid, gid); }
    setreuid (ruid, euid);
    return ec;
}

/* xChmod (filehandle, mode)
**     setzt die Zugriffsrechte fr die Folder- und Indexdatei auf den
**     angegebenen Wert `mode'.
*/
int xChmod (xFile fx, int mode)
{
    int ec = fchmod (fx->dfd, mode);
    if (ec < 0) { return ec; }
    return fchmod (fx->xfd, mode);
}

size_t xGetExtra (xFile fx, void *buf, size_t sz)
{
    if (sz > fx->ixlen - OFFSIZE) { sz = fx->ixlen - OFFSIZE; }
    if (sz > 0) { memcpy (buf, fx->extra + OFFSIZE, sz); }
    return sz;
}

size_t xSetExtra (xFile fx, void *buf, size_t sz)
{
    ifnull (strchr ("iuw", fx->mode[0])) { errno = EACCES; return -1; }
    if (sz > fx->ixlen - OFFSIZE) { sz = fx->ixlen - OFFSIZE; }
    if (sz > 0) { memcpy (fx->extra + OFFSIZE, buf, sz); }
    return sz;
}

size_t xGetPrefix (xFile fx, void *buf, size_t sz)
{
    if (sz > fx->ixstart - 2*sizeof(off_t)) {
	sz = fx->ixstart - 2*sizeof(off_t);
    }
    if (sz > 0) { memcpy (buf, fx->prefix + 2*sizeof(off_t), sz); }
    return sz;
}

size_t xSetPrefix (xFile fx, void *buf, size_t sz)
{
    ifnull (strchr ("iuw", fx->mode[0])) { errno = EACCES; return -1; }
    if (sz > fx->ixstart - 2*sizeof(off_t)) {
	sz = fx->ixstart - 2*sizeof(off_t);
    }
    if (sz > 0) { memcpy (fx->prefix + 2*sizeof(off_t), buf, sz); }
    fx->pfx_saved = 0;
    return sz;
}

char xAmode (xFile fx)
{
    return fx->mode[0];
}

void xCopyPrefix (xFile fx, xFile fy, size_t sz)
{
    char *buf = galloc (__FILE__, __LINE__, sz);
    xGetPrefix (fx, buf, sz);
    xSetPrefix (fy, buf, sz);
    fy->pfx_saved = 0;
    gfree (buf);
}

void xCopyExtra (xFile fx, xFile fy, size_t sz)
{
    char *buf = galloc (__FILE__, __LINE__, sz);
    xGetExtra (fx, buf, sz);
    xSetExtra (fy, buf, sz);
    gfree (buf);
}

int xStat (xFile fx, struct stat *buf)
{
    return fstat (fx->dfd, buf);
}

/* Prfe, ob Datendatei und/oder Indexdatei existieren und gib das das Ergebnis
** als Wert im Bereich 0 (Weder Daten- noch Indexdatei existieren) bis 3
** (beide Dateien existieren) zurck. Das untere Bit steht hierbei fr
** die (Nicht-)Existenz der Datendatei, das obere fr die Indexdatei.
*/
int xCheck (char *file)
{
    struct stat ds, xs;
    int rd, rx;
    char *ix = gconc (__FILE__, __LINE__, file, IXSFX, NULL);
    rd = stat (file, &ds) ? 0 : 1;
    rx = stat (file, &xs) ? 0 : 2;
    gfree (ix);
    return rd|rx;
}

/* xSize (fx) - gibt die aktuelle Lnge der Folderdatei zurck. */
off_t xSize (xFile fx)
{
    off_t sav, res;
    xFlush (fx);
    sav = lseek (fx->dfd, 0, SEEK_CUR);
    res = lseek (fx->dfd, 0, SEEK_END);
    (void) lseek (fx->dfd, sav, SEEK_SET);
    return res;
}

int xRename (char *old, char *new)
{
    char *idxold = gconc (__FILE__, __LINE__, old, IXSFX, NULL),
	 *idxnew = gconc (__FILE__, __LINE__, new, IXSFX, NULL);
    xFile fx = xOpen (old, "rx");
    int res = -1;
    ifnull (fx) { goto xRename_result; }
    if (rename (old, new)) { goto xRename_close; }
    if (rename (idxold, idxnew)) { rename (new, old); goto xRename_close; }
    res = 0;
xRename_close:
    xClose (fx);
xRename_result:
    gfree (idxold); gfree (idxnew);
    return res;
}

int xMsgNo (xFile fx)
{
    return fx->msgno;
}
